/*!
    \file   prng.cpp
    \brief  Adapted for C++ from Mersenne-Twister code by T. Nishimura and M.
            Matsumoto. The original code is licensed as follows:

    <pre>
    Copyright (C) 2004, Makoto Matsumoto and Takuji Nishimura,
    All rights reserved.                          
 
    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:
 
      1. Redistributions of source code must retain the above copyright
         notice, this list of conditions and the following disclaimer.
 
      2. Redistributions in binary form must reproduce the above copyright
         notice, this list of conditions and the following disclaimer in the
         documentation and/or other materials provided with the distribution.
 
      3. The names of its contributors may not be used to endorse or promote 
         products derived from this software without specific prior written 
         permission.
 
    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
    A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
    CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
    EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
    PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
    PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
    NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
    SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
    References:
    T. Nishimura, ``Tables of 64-bit Mersenne Twisters''
      ACM Transactions on Modeling and 
      Computer Simulation 10. (2000) 348--357.
    M. Matsumoto and T. Nishimura,
      ``Mersenne Twister: a 623-dimensionally equidistributed
        uniform pseudorandom number generator''
      ACM Transactions on Modeling and 
      Computer Simulation 8. (Jan. 1998) 3--30.
 
    Any feedback is very welcome.
    http://www.math.hiroshima-u.ac.jp/~m-mat/MT/emt.html
    email: m-mat @ math.sci.hiroshima-u.ac.jp (remove spaces)
    </pre>
 */
#if defined (HAVE_STDINT_H)
#   include <cstdint>
#else
#   include <inttypes.h>
#endif
#include <cmath>
#include <sys/time.h>

#include "prng.h"


#define NN       312
#define MM       156
#define MATRIX_A UINT64_C(0xB5026F5AA96619E9)
#define UM       UINT64_C(0xFFFFFFFF80000000) /* Most significant 33 bits */
#define LM       UINT64_C(0x7FFFFFFF) /* Least significant 31 bits */

namespace lsst {

namespace {

/* The array for the state vector */
static uint64_t mt[NN]; 

/* mti==NN+1 means mt[NN] is not initialized */
static int mti = NN + 1; 


/* Initializes mt[NN] with a seed */
static void initRand64(uint64_t seed)
{
    mt[0] = seed;
    for (mti = 1; mti < NN; mti++)
    {
        mt[mti] = UINT64_C(6364136223846793005) * (mt[mti-1] ^ (mt[mti-1] >> 62)) + mti;
    }
}


/* Initializes with an array of seeds */
static void initRand64(uint64_t * keys, uint64_t nkeys)
{
    uint64_t i, j, k;
    initRand64(UINT64_C(19650218));
    i = 1;
    j = 0;
    k = (NN > nkeys ? NN : nkeys);
    for (; k; k--)
    {
        // non linear
        mt[i] = (mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 62)) * UINT64_C(3935559000370003845))) + keys[j] + j;
        i++;
        j++;
        if (i >= NN)
        {
            mt[0] = mt[NN - 1];
            i = 1;
        }
        if (j >= nkeys)
        {
            j = 0;
        }
    }
    for (k = NN - 1; k; k--)
    {
        // non linear
        mt[i] = (mt[i] ^ ((mt[i - 1] ^ (mt[i - 1] >> 62)) * UINT64_C(2862933555777941757))) - i;
        i++;
        if (i >= NN)
        {
            mt[0] = mt[NN - 1];
            i = 1;
        }
    }

    mt[0] = UINT64_C(1) << 63; /* MSB is 1; assuring non-zero initial array */ 
}


/* Generates a random number in the interval [0, 2^64 - 1] */
static uint64_t genRand64()
{
    static uint64_t const mag01[2] = { 0, MATRIX_A };

    int i;
    uint64_t x;

    if (mti >= NN) 
    {
        // generate NN words at a time

        // if initRand64() has not been called
        // a default initial seed is used
        if (mti == NN + 1)
        {
            initRand64(UINT64_C(5489)); 
        }
        for (i = 0; i < NN - MM; i++)
        {
            x = (mt[i] & UM) | (mt[i + 1] & LM);
            mt[i] = mt[i + MM] ^ (x >> 1) ^ mag01[static_cast<int>(x & UINT64_C(1))];
        }
        for (;i<NN-1;i++) {
            x = (mt[i] & UM) | (mt[i + 1] & LM);
            mt[i] = mt[i + (MM - NN)] ^ (x >> 1) ^ mag01[static_cast<int>(x & UINT64_C(1))];
        }
        x = (mt[NN - 1] & UM) | (mt[0] & LM);
        mt[NN - 1] = mt[MM - 1] ^ (x >> 1) ^ mag01[static_cast<int>(x & UINT64_C(1))];

        mti = 0;
    }
  
    x = mt[mti++];

    x ^= (x >> 29) & UINT64_C(0x5555555555555555);
    x ^= (x << 17) & UINT64_C(0x71D67FFFEDA60000);
    x ^= (x << 37) & UINT64_C(0xFFF7EEE000000000);
    x ^= (x >> 43);

    return x;
}

}    // end of anonymous namespace


void initRandom()
{
    uint64_t init[4] = { UINT64_C(0x12345),
                         UINT64_C(0x23456), 
                         UINT64_C(0x34567),
                         UINT64_C(0x45678) };    
    struct timeval t;
    gettimeofday(&t, 0);
    init[0] += static_cast<uint64_t>(t.tv_sec);
    init[2] += static_cast<uint64_t>(t.tv_usec);

    initRand64(init, 4);
}

double random()
{
    return (genRand64() >> 11) * (1.0/9007199254740992.0);
}

double random2()
{
    return (genRand64() >> 11) * (1.0/9007199254740991.0);
}

double normal()
{
    // Use Box-Muller transform to get a normal distribution
    // from the uniform distribution
    double x, y, s;

    do
    {
        x = 2.0*random2() - 1.0;
        y = 2.0*random2() - 1.0;
        s = x*x + y*y;
    }
    while (s > 1.0 || s == 0.0);
    // Even though s == 0.0 is extremely unlikely, check for it and
    // discard (since it will produce a NaN below)
    return x * std::sqrt((- 2.0*std::log(s)) / s);
}

bool coinToss(double const p)
{
    return (random() <= p);
}

}   // end of namespace lsst
